﻿// Copyright 2010 Canonical Ltd.
// 
// This file is part of UbuntuOne on Windows.
// 
// UbuntuOne on Windows is free software: you can redistribute it and/or modify		
// it under the terms of the GNU Lesser General Public License version 		
// as published by the Free Software Foundation.		
//		
// Ubuntu One on Windows is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	
// GNU Lesser General Public License for more details.	
//
// You should have received a copy of the GNU Lesser General Public License	
// along with UbuntuOne for Windows.  If not, see <http://www.gnu.org/licenses/>.
// 
// Authors: Manuel de la Peña <manuel.delapena@canonical.com>
using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Web;
using System.Windows;
using Canonical.Ubuntu.SSO;
using Canonical.Ubuntu.SSO.Service;
using Canonical.Ubuntu.SSO.Views;
using Canonical.UbuntuOne.Client.Preferences;
using Canonical.UbuntuOne.Common.Aop;
using Canonical.UbuntuOne.Common.Utils;
using log4net;
using Newtonsoft.Json.Linq;
using OAuth;

namespace Canonical.UbuntuOne.Client.Notification
{
    /// <summary>
    /// Implementation of the INotificationIconPresenter interface that manages a view that is used
    /// to display notifications to the user.
    /// </summary>
    public class NotificationIconPresenter : INotificationIconPresenter
    {
        #region Variables

        /// <summary>
        /// The url to ebe openened when the user wants to upgrade.
        /// </summary>
        public const string UpgradeUrl = "https://one.ubuntu.com/upgrade";

        /// <summary>
        /// The user to be openened when the user requires help.
        /// </summary>
        public const string HelpUrl = "https://one.ubuntu.com/support/";

        private static readonly string SharesLocation = "";
        private static readonly string SyncsLocation = "";
        private ISyncDaemonClientNotifier _clientNotifier;
        private ILog _logger;
        private readonly object _loggerLock = new object();

        #endregion

        #region Di properties

        /// <summary>
        /// Gets and sets the dialog used to show the preferences.
        /// </summary>
        public IPreferencesDialogPresenter PreferencesDialogPresenter { get; set; }

        /// <summary>
        /// Gets and sets the views that will be used to show the user the information.
        /// </summary>
        public INotificationIconView NotificationIconView { get; set; }

        /// <summary>
        /// Gets and sets the mapper used to convert from the state enum to a string.
        /// </summary>
        public IStateMapper StateMapper { get; set; }

        /// <summary>
        /// Gets and sets the webbrowser to be used to show pages to the user.
        /// </summary>
        public IWebbrowser Webbrowser { get; set; }

        /// <summary>
        /// Get and sets the eplorer that will be used to open folder in the client 
        /// machine.
        /// </summary>
        public IExplorer Explorer { get; set; }

        /// <summary>
        /// Gets and sets the application that is running the notify presenter.
        /// </summary>
        public IApplication Application { get; set; }

        /// <summary>
        /// Gets and sets the notifier that will ensure that the presenter know about changes in the 
        /// sync daemon.
        /// </summary>
        public ISyncDaemonClientNotifier SyncDaemonClientNotifier
        {
            get { return _clientNotifier; }
            set
            {
                if (value == null) return;
                _clientNotifier = value;
                // register to the different events raised by the client
                _clientNotifier.OnNotification += OnNotificationHandler;
                _clientNotifier.OnOperationCompletness += OnOperationCompletnessHandler;
                _clientNotifier.OnRecentlyModifiedChanged += OnRecentlyModifiedChangedHandler;
                _clientNotifier.OnStateChanged += OnStateChangedHandler;
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets and sets the logger used by the class.
        /// </summary>
        internal ILog Logger
        {
            get
            {
                if (_logger == null)
                {
                    lock (_loggerLock)
                    {
                        _logger = LogManager.GetLogger(typeof(NotificationIconPresenter));
                    }
                }
                return _logger;
            }
            set { _logger = value; }
        }

        #endregion

        #region Constructors

        static NotificationIconPresenter()
        {
            SharesLocation = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            SyncsLocation = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
        }

        #endregion

        #region Helper methods

        public HttpWebRequest MakeRequest(string uri, string consumerKey, string consumerSecret, string toke, string tokenSecret)
        {
            // Form the full REST request url
            Uri url = new Uri(uri);

            // Instantiate OAuthBase and declare variables
            var oAuth = new OAuthBase();
            var nonce = oAuth.GenerateNonce();
            var timeStamp = oAuth.GenerateTimeStamp();
            var normUrl = string.Empty;
            var normParams = string.Empty;
            var strRequest = string.Empty;

            // Create an OAuth signature
            string signature = oAuth.GenerateSignature(url,
                consumerKey, consumerSecret, toke, tokenSecret,
                "GET", timeStamp, nonce, OAuthBase.SignatureTypes.HMACSHA1,
                out normUrl, out normParams);

            // Construct the OAuth authenticated REST url
            strRequest = normParams + "&" + HttpUtility.UrlEncode("oauth_signature") + "=" + HttpUtility.UrlEncode(signature);
            strRequest = strRequest.Replace("&", @""",");
            strRequest = strRequest.Replace("=", @"=""");
            strRequest += @"""";
            strRequest = @"OAuth realm=""""," + strRequest;
            var request = WebRequest.Create(normUrl) as HttpWebRequest;
            request.Headers.Add("Authorization", strRequest);
            return request;
        }

        /// <summary>
        /// Helper method that will ensure that when the state of the sync daemon changes this change
        /// is correctly reflected in the View.
        /// </summary>
        /// <param name="sender">The object that sent the event.</param>
        /// <param name="e">The arguments that describe the state change.</param>
        [DebugLogged]
        private void OnStateChangedHandler(object sender, StateEventArgs e)
        {
            NotificationIconView.State = StateMapper.ToString(e.State);
        }

        /// <summary>
        /// Helper method that will ensure that when the recently changed files are updated by the daemon that 
        /// such a change is correcty displayed in the view.
        /// </summary>
        /// <param name="sender">The object that sent the event.</param>
        /// <param name="e">The arguments that describe the changes in the recently changed list.</param>
        [DebugLogged]
        private void OnRecentlyModifiedChangedHandler(object sender, RecentlyModifiedEventArgs e)
        {
            NotificationIconView.RecentlyModifiedFiles = e.RecentlyModifiedFiles;
        }

        /// <summary>
        /// Helper method that will ensure that when the status of an operation that is being performed by the daemon
        /// changes that such a changed ins correctly reflected in the UI.
        /// </summary>
        /// <param name="sender">The object that sent the event.</param>
        /// <param name="e">The argumets that describe the new updated operation state.</param>
        [DebugLogged]
        private void OnOperationCompletnessHandler(object sender, OperationCompletenessEventArgs e)
        {
            NotificationIconView.CurrentOperation = String.Format(NotificationResources.OperationStateNotification,
                                                                  e.Operation, e.Percentage, e.Speed, e.SpeedUnits);
        }

        /// <summary>
        /// Helper method that will ensure that when a notification is sent by the daemon that it is correctly displayed in the UI.
        /// </summary>
        /// <param name="sender">The object that sent the event.</param>
        /// <param name="e">The arguments that contains the notification to display.</param>
        [DebugLogged]
        private void OnNotificationHandler(object sender, NotificationEventArgs e)
        {
            NotificationIconView.Notification = e.Message;
        }

        #endregion

        #region Implementation of INotificationIconPresenter

        /// <summary>
        /// When implemented this method should for the sync of the ubuntu one folder.
        /// </summary>
        public void ManualSync()
        {
            // TODO: use the oauth code from the provider not here!
            var keyring = new Keyring();
            keyring.DataProtector = new DPAPIDataProtector();
            var secret = keyring.GetSecretByName("Default", "UbuntuOne");
            if (secret == null)
            {
                var dialog = new LoginDialog();
                dialog.ShowDialog();
                if (dialog.WasCanceled)
                {
                    MessageBox.Show("You need to login to be able to sync your files.");
                    return;
                }
                try
                {
                    var auth = new Authentications();
                    var credentials = JObject.Parse(auth.Authenticate(dialog.Email, dialog.Password, "Ubuntu One @ Windows"));
                    secret = string.Format("{0}:{1}:{2}:{3}", 
                        (string)credentials["token"],
                        (string)credentials["token_secret"],
                        (string)credentials["consumer_key"],
                        (string)credentials["consumer_secret"]);
                    // TODO: Move this out of here!!! put it in the correct location, like the SSOProvider!!
                    // ping the service to make it download the tokens
                    var pingRequest = MakeRequest(Constants.PingUrl + dialog.Email, (string) credentials["consumer_key"],
                                                  (string) credentials["consumer_secret"],
                                                  (string) credentials["token"], (string) credentials["token_secret"]);
                    var pingRepomse = pingRequest.GetResponse();
                    keyring.CreateSecret("Default", "UbuntuOne", secret);
                }
                catch (Exception e)
                {
                    MessageBox.Show("You provided the wrong credentials.");
                    return;
                }
            }

            // TODO: Use the service rather than do it here.
            // TODO: This has to be fixed, we should not call the command directly, user the
            // process dispatcher to prepare for the syncdaemon!!!!
            // TODO: Do not hardcode the folder, use an object to store that

            var oneDir = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments),
                "UbuntuOne");
            // ensure that the path can be handled by the python code
            oneDir = oneDir.Replace("\\", "\\\\");
            if (!Directory.Exists(oneDir))
            {
                Directory.CreateDirectory(oneDir);
            }

            if (!Directory.Exists(Path.Combine(oneDir, ".ubuntuone-sync")))
            {
                // init the directory
                var init = new Process
                {
                    StartInfo =
                    {
                        WorkingDirectory = @"C:\Program Files\Canonical\UbuntuOne\Client\U1Sync",
                        FileName = @"C:\Program Files\Canonical\UbuntuOne\Client\U1Sync\main.exe",
                        Arguments = String.Format(" --init \"{0}\" --oauth {1}",
                            oneDir, secret),
                        UseShellExecute = false,
                        RedirectStandardOutput = false,
                        RedirectStandardError = false,
                        CreateNoWindow = true,
                        WindowStyle = ProcessWindowStyle.Hidden
                    }
                };
                init.Start();
                init.WaitForExit();
                init.Close();
            }

            var proc = new Process
            {
                StartInfo =
                {
                    WorkingDirectory = @"C:\Program Files\Canonical\UbuntuOne\Client\U1Sync",
                    FileName = @"C:\Program Files\Canonical\UbuntuOne\Client\U1Sync\main.exe",
                    Arguments = String.Format("\"{0}\" --oauth {1}",
                        oneDir, secret),
                    UseShellExecute = false,
                    RedirectStandardOutput = false,
                    RedirectStandardError = false,
                    CreateNoWindow = true,
                    WindowStyle = ProcessWindowStyle.Hidden
                }
            };
            proc.Start();

        }

        /// <summary>
        /// Opens the Preferences window of Ubuntu One.
        /// </summary>
        public void OpenUbuntuOnePreferences()
        {
            PreferencesDialogPresenter.Show();
        }

        /// <summary>
        /// Opens the explorer with the location of the shares of Ubuntu One.
        /// </summary>
        public void OpenSharesLocation()
        {
            try
            {
                Explorer.OpenFolder(SharesLocation);
            }
            catch (ExplorerException exception)
            {
                Logger.WarnFormat("There was an error operning the SharesLocation, {0}", exception);
                // use the view to notify the error
                NotificationIconView.Notification = String.Format(NotificationResources.ErrorOpeningFolder,
                    Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments));
                throw;
            }
        }

        /// <summary>
        /// Opens the explorer with the location of the sync folders.
        /// </summary>
        public void OpenSyncsLocation()
        {
            try
            {
                Explorer.OpenFolder(SyncsLocation);
            }
            catch (ExplorerException exception)
            {
                Logger.WarnFormat("There was an error operning the SharesLocation, {0}", exception);
                // use the view to notify the error
                NotificationIconView.Notification = String.Format(NotificationResources.ErrorOpeningFolder,
                    Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments));
                throw;
            }
        }

        /// <summary>
        /// Opens the location used to provide help to the user.
        /// </summary>
        public void OpenHelp()
        {
            Webbrowser.LauncWebpage(HelpUrl);
        }

        /// <summary>
        /// Opens the location used to upgrade the account of the user.
        /// </summary>
        public void OpenUpgradeOption()
        {
            Webbrowser.LauncWebpage(UpgradeUrl);
        }

        /// <summary>
        /// Close the client and the daemon.
        /// </summary>
        public void Exit()
        {
            Application.Exit();
        }

        #endregion
    }
}
