﻿/*
 * Copyright 2010 Canonical Ltd.
 * 
 * This file is part of UbuntuOne on Windows.
 * 
 * UbuntuOne on Windows is free software: you can redistribute it and/or modify		
 * it under the terms of the GNU Lesser General Public License version 		
 * as published by the Free Software Foundation.		
 *		
 * Ubuntu One on Windows is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	
 * GNU Lesser General Public License for more details.	
 *
 * You should have received a copy of the GNU Lesser General Public License	
 * along with UbuntuOne for Windows.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Authors: Manuel de la Peña <manuel.delapena@canonical.com>
 */
using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using Canonical.UbuntuOne.Common;
using Canonical.UbuntuOne.Common.OperationContracts;
using log4net;

namespace Canonical.UbuntuOne.ProcessDispatcher
{
    /// <summary>
    /// Implementation of the IEventNotifier that receives messages from the python code through sockets and 
    /// broadcasts those messages to the .net clients that have registered to receive them.
    /// </summary>
    internal class EventNotifier : IEventNotifier
    {
        #region Helper class

        /// <summary>
        /// This class is used to allows to keep the state of the async reading operation of the 
        /// socket.
        /// </summary>
        private class StateObject
        {
            /// <summary>
            /// The socket we are working with.
            /// </summary>
            public Socket WorkSocket;

            /// <summary>
            /// The size of the buffer that will be used to read.
            /// </summary>
            public const int BufferSize = 1024;

            /// <summary>
            /// The buffer wich will be used to place the read information.
            /// </summary>
            public readonly byte[] Buffer = new byte[BufferSize];

            /// <summary>
            /// Stream that is used to concatenated the data that has been processed
            /// so far.
            /// </summary>
            public readonly MemoryStream ResultBuilder = new MemoryStream();
        }

        #endregion

        #region Vars

        private ILog _logger;
        private readonly object _loggerLock = new object();
        private readonly object _clientsLock = new object();
        private readonly object _portManagerLock = new object();
        private readonly object _isListeningLock = new object();
        private static readonly HashSet<ISyncDaemonClient> _registeredClients = new HashSet<ISyncDaemonClient>();
        private bool _isListening;

        #endregion

        #region DI properties

        /// <summary>
        /// Gets and sets the object that is used to retrieve the information of the object
        /// that performs the different WCF calls.
        /// </summary>
        internal ICallerContext<ISyncDaemonClient> CallerContext { get; set; }

        /// <summary>
        /// Gets and sets the logger that is used to log the different activities.
        /// </summary>
        internal ILog Logger
        {
            get
            {
                if(_logger == null)
                {
                    lock (_loggerLock)
                    {
                        _logger = LogManager.GetLogger(typeof (EventNotifier));
                    }
                }
                return _logger;
            }
            set { _logger = value; }
        }


        #endregion

        #region Implementation of IEventNotifier

        /// <summary>
        /// Method that allows to register a client that will be reciving notitications 
        /// from than point until it unregisters
        /// </summary>
        public void RegisterClient()
        {
            lock (_clientsLock)
            {
                // we get the client that will be used for the callbacks
                var client = CallerContext.GetCaller();
                if(!_registeredClients.Contains(client))
                {
                    _registeredClients.Add(client);
                }
            }
        }

        /// <summary>
        /// Method that allows to unregister a client from receiving notification.
        /// </summary>
        public void UnregisterClient()
        {
            lock (_clientsLock)
            {
                var client = CallerContext.GetCaller();
                if(_registeredClients.Contains(client))
                {
                    _registeredClients.Remove(client);
                }
            }
        }

        /// <summary>
        /// Method that tells the notifier to start listening for other events using a separate thread.
        /// </summary>
        public void StartListening()
        {
            // TODO: Implement when ever a new techinque is used
            throw new NotImplementedException();
        }

        /// <summary>
        /// Method that tells the notifier to stop listning for events.
        /// </summary>
        public void StopListening()
        {
            // TODO: Implement whenever a new techique is used.
        }

        /// <summary>
        /// Gets if the notifier is listening.
        /// </summary>
        public bool IsListening
        {
            get { return _isListening; }
        }

        #endregion
    }
}
